#!/usr/bin/env python
""" Used to suppress or block a regularly-scheduled or daemonized job from
    running or signal it to suspend or suppress operations.

    Cletus_supp is highly opinionated about scheduling:
       - Sometimes schedules need to be suppressed or suspended - maybe to
         prevent concurrency problems when a backup is running, maybe to
         reduce the number of jobs running to help diagnose a problem, etc.
       - Most teams rely on commenting jobs out in cron, or killing them,
         or modifying their config to do this.  These approaches are
         error-prone and often don't work well at all.
       - What's needed is the ability to temporarily suppress a schedule.

    That's what this module does - in a way that makes it easy for the apps:
       - Every application gets a suppression directory.  It can be
            - Either explicitely provided
            - Or left to default to XDG config standard.  On linux this
              would be: /home/.config/<appname>/suppression
         If the directory does not exist it will be created.
       - Applications can be "suppressed" by just touching a file or otherwise
         writing a file in this dir that complies with the following
         naming convention.
       - Suppression file naming convention:
            - name-<name>.suppress
            - <name> can be used by the application to suppress specific
              configurations without requiring dedicated directories for each.
              For example, if an application supported 400 customer accounts,
              rather than create 400 dedicated suppression directories, they
              could share a single suppression directory and use the name to
              identify which is being suppressed.
            - name-all.suppress will suppress all names.
        - Applications can check to see if they are suppressed at the begining
          of their operations, or at checkpoints.  If they're suppressed they
          can do whatever is desired - shut down gracefully, sleep, etc.

    Example Usage Scenario:
       - Large application server runs a hundred processes for extracting,
         aggregating, loading, validating, analyzing, and maintaining data.
       - Most processes run as daemons polling for the next set of work, some
         are scheduled to run at specific times (such as backups).
       - In order to minimize the number of corrupted or partial files backed-up,
         the backup process as an initial step that creates a suppression file
         for some jobs, then sleeps for 15 minutes prior to the actual backup
         starting.   When the backup is complete it removes this file.
       - One of the jobs that is affected runs as a daemon, and gets a new
         file to process every 10 minutes.  Every minute it polls for this file,
         but before it polls it checks to see if it's suppressed.  If it is, then
         it sleeps for another minute instead.
       - Another of the jobs creates a report once a day and emails it to admins.
         This job first runs a routine to wait until it's not suspended.  It will
         just keep sleeping then checking every ten minutes until it is no longer
         suspended.

    To Dos:
       - Allow use of ZooKeeper & database as an alternative to the config
         directory.  This will better support distributed processing.
       - Add more key-value pairs to suppression file name in order to support:
            - suppression start times
            - suppression stop times
            - suppression durations
            - suppression count
            - identity of suppressor
       - Add a suppression-file editor feature to help apps read & write the
         files.

    See the file "LICENSE" for the full license governing use of this file.
    Copyright 2013, 2014 Ken Farmer
"""


import os
import errno
import glob
import logging

import appdirs



class NullHandler(logging.Handler):
    def emit(self, record):
        #print record
        pass


class SuppressCheck(object):
    """ Typical Usage:
           supp_check = mod.SuppressCheck(pgm_name)
           if supp_check.suppressed():
               print 'WARNING: filemover has been suppressed'
               sys.exit(0)
        Inputs
           - app_name   - Used to look up XDG config dir unless config_dir
                          is specifically provided.  Also used for suppress
                          file name if that is not provided.
           - log_name   - Defaults to '__main__'
           - config_dir - If not provided will use XDG dirctory with app_name.
                          Defaults to None.
        Raises
           - OSError if config_dir doesn't exist and cannot be
             created or accessed.
    """

    def __init__(self,
                 app_name,
                 log_name='__main__',
                 config_dir=None):
        # set up logging
        self.logger   = logging.getLogger('%s.cletus_supp' % log_name)
        # don't print to sys.stderr if no parent logger has been set up:
        #logging.getLogger(log_name).addHandler(logging.NullHandler())
        self.logger.debug('SuppressCheck starting now')

        self.app_name        = app_name
        if config_dir:
            self.config_dir  = os.path.join(config_dir, 'suppress')
            self.logger.debug('config_dir derrived from arg: %s' % self.config_dir)
        else:
            self.config_dir  = os.path.join(appdirs.user_config_dir(app_name), 'suppress')
            self.logger.debug('config_dir provided via user_config_dir: %s' % self.config_dir)

        try:
            os.makedirs(self.config_dir)
            self.logger.info('Suppression dir created successfully')
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                self.logger.critical('Unknown OSError on creating config dir')
                raise




    def suppressed(self, suppress_name=None):
        """ Determines from suppress_name whether or not that process is
            suppressed.  Returns True/False.
            - if suppress_name is provided, then it can be suppressed by either:
                - a suppress file of name-<suppress_name>.suppress
                - or a suppress file of name-all.suppress
            - if suppress is not provided, then it defaults to the app_name
        """
        self.names_suppressed = self._get_suppressed_names()

        if suppress_name is None:
            suppress_name = self.app_name

        if 'name-all.suppress' in self.names_suppressed:
            self.logger.info('Process has been suppressed')
            return True
        else:
            full_suppress_name = 'name-%s.suppress' % suppress_name
            if full_suppress_name in self.names_suppressed:
                self.logger.info('Process has been suppressed')
                return True
            else:
                self.logger.debug('Process has NOT been suppressed')
                return False


    def _get_suppressed_names(self):
        raw_files   = glob.glob(os.path.join(self.config_dir, '*.*'))
        clean_files = []
        for one_file in raw_files:
            self.logger.debug('checking file: %s' % one_file)
            if _valid_suppress_file(one_file):
                head, tail = os.path.split(one_file)
                clean_files.append(tail)
            else:
                self.logger.critical('invalid suppress file: %s' % one_file)
                raise ValueError
        return clean_files


def _valid_suppress_file(name):
    head, tail = os.path.split(name)
    file_name, file_extension = os.path.splitext(tail)
    if file_extension != '.suppress':
        return False
    elif not file_name.startswith('name-'):
        return False
    else:
        return True







#!/usr/bin/env python
""" Used to manage access to config files.

    Cletus_config is highly opinionated about what constitutes good config
    management:
       - Config files should be kept within the xdg directory:
            - linux:  $HOME/.config/<app name>
       - But overrides to the location, for special runs, migrations,
         testing, etc are occasionally necessary.
       - The best format at present for config files is yaml.
       - Config file contents should be validated.
       - Arguments should override config items.
       - It's easier to reference config items as namespaces than
         dictionaries.

    The objective of cletus_config is to deliver on this optinions in a way
    that makes it easy for applications.

    See the file "LICENSE" for the full license governing use of this file.
    Copyright 2013, 2014 Ken Farmer
"""

# todo:
# 1. validate the validation schema, sigh, well at least a little
# 2. allow users to add args and override config with them
# 3. allow users to lookup environ variables and override config with them

from __future__ import division
import os
import sys

import argparse
import time
import logging
from pprint import pprint as pp

import yaml
import validictory as valid
import appdirs


class NullHandler(logging.Handler):
    def emit(self, record):
        #print record
        pass


class ConfigManager(object):

    def __init__(self,
                 config_schema=None,
                 log_name='__main__',
                 namespace_access=True):

        self.cm_namespace_access = namespace_access

        # set up logging:
        self.cm_logger   = logging.getLogger('%s.cletus_config' % log_name)
        # don't print to sys.stderr if no parent logger has been set up:
        # logging.getLogger(log_name).addHandler(logging.NullHandler())
        self.cm_logger.debug('ConfigManager starting now')

        self.cm_config_schema    = config_schema
        self.cm_config_fqfn      = None

        self.cm_config_file      = {}
        self.cm_config_env       = {}
        self.cm_config_namespace = {}
        self.cm_config_iterable  = {}
        self.cm_config_defaults  = {}
        self.cm_config           = {}

        # store an original copy of variable names to use to protect
        # from updating by bunch later on.
        self.cm_orig_dict_keys   = self.__dict__.keys()


    def _bunch(self):
        for key, val in self.cm_config.items():
            if key in self.cm_orig_dict_keys:
                raise ValueError, 'config key is a reserved value: %s' % key
            elif key in dir(ConfigManager):
                raise ValueError, 'config key is a reserved value: %s' % key
            else:
                self.__dict__[key] = val

    def _post_add_maintenance(self, config):
        clean_config = self._remove_null_overrides(config)
        self.cm_config.update(clean_config)

        self.log_level = self.cm_config.get('log_level', None)
        if self.cm_namespace_access:
            self._bunch()


    def _remove_null_overrides(self, dictname):
        working_dict = dict(dictname)
        for key in dictname:
            if dictname[key] is None and self.cm_config.get(key, None) is not None:
                working_dict.pop(key)
        return working_dict


    def add_file(self,
                 app_name=None,
                 config_dir=None,
                 config_fn=None,
                 config_fqfn=None):

        # figure out the config_fqfn:
        if config_fqfn:
            self.cm_config_fqfn = config_fqfn
            self.cm_logger.debug('using config_fqn: %s' % config_fqfn)
        elif config_dir and config_fn:
            self.cm_config_fqfn = os.path.join(config_dir, config_fn)
            self.cm_logger.debug('using config_dir & config_fn: %s' % config_fqfn)
        elif app_name and config_fn:
            self.cm_config_fqfn = os.path.join(appdirs.user_config_dir(app_name), config_fn)
            self.cm_logger.debug('using app_name & config_fn: %s' % config_fqfn)
        else:
            self.cm_logger.critical('Invalid combination of args.  Provide either config_fqfn, config_dir + config_fn, or app_name + config_fn')
            raise ValueError, 'invalid config args'

        if not os.path.isfile(self.cm_config_fqfn):
            self.cm_logger.critical('config file missing: %s' % self.cm_config_fqfn)
            raise IOError, 'config file missing, was expecting %s' % self.cm_config_fqfn

        self.cm_config_file = {}
        with open(self.cm_config_fqfn, 'r') as f:
            self.cm_config_file = yaml.safe_load(f)

        self._post_add_maintenance(self.cm_config_file)



    def add_env_vars(self, key_list=None, key_to_lower=False):
        assert key_to_lower in [True, False]
        self.cm_config_env = {}

        final_key_list = key_list or self._get_schema_keys()
        if not final_key_list:
            raise ValueError, 'add_env_vars called without key_list or cm_config_schema'

        # get list of tuples of environment variables
        env_list = os.environ.items()

        for env_tup in env_list:
            if env_tup[0] in final_key_list:
                if key_to_lower:
                    self.cm_config_env[env_tup[0].lower()] = env_tup[1]
                else:
                    self.cm_config_env[env_tup[0]] = env_tup[1]

        self._post_add_maintenance(self.cm_config_env)


    def _get_schema_keys(self):
        if self.cm_config_schema:
            keylist = [var.upper() for var in self.cm_config_schema['properties'].keys()]
            return keylist
        else:
            return []


    def add_namespace(self, args):
        self.cm_config_namespace = {}
        self.cm_config_namespace.update(vars(args))
        self._post_add_maintenance(self.cm_config_namespace)

    def add_iterable(self, user_iter):
        self.cm_config_iterable = {}
        self.cm_config_iterable.update(user_iter)
        self._post_add_maintenance(self.cm_config_iterable)

    def add_defaults(self, default_dict):
        """ Applies defaults to empty config items with the following limits:
            - only if they have a default field created within the schema
            - only if they are top-level items - no nested items
        """
        # first create dict with all needed defaults:
        for key in self.cm_config:
           if self.cm_config[key] is None and default_dict.get(key, None) is not None:
               self.cm_config_defaults[key] = default_dict[key]

        # next update the main config from it:
        self._post_add_maintenance(self.cm_config_defaults)



    def validate(self, config_type='config', schema=None):

        config_schema = schema or self.cm_config_schema

        config_types = {'config':           self.cm_config,
                        'config_file':      self.cm_config_file,
                        'config_env':       self.cm_config_env,
                        'config_namespace': self.cm_config_namespace,
                        'config_iterable':  self.cm_config_iterable}
        assert config_type in config_types

        if self.cm_config_schema:
            try:
                valid.validate(config_types[config_type], config_schema)
            except valid.FieldValidationError as e:
                self.cm_logger.critical('Config error on field %s' % e.fieldname)
                self.cm_logger.critical(e)
                raise ValueError, 'config error: %s' % e
            else:
                return True
        else:
            return False






#!/usr/bin/env python
""" Used to quickly and easily set up somewhat comprehensive logging.

    Cletus_log is highly opinionated about what constitutes good logging:
       - Logs must be formatted with date, module name, level and msg
         so they can be sorted, searched, or reported on.
       - Logs should be written to a local file system, and should be
         automatically rotated to avoid filling the file system.
       - Logs should be written to the console as well for easy
         interactive debugging.
       - Logs should be trivial to set up, have intelligent defaults,
         be easy to maintain consistency across multiple programs,
         and expose as little logging jargon as possible.

    Anything else is a distraction, is insufficient, or is a (possibly
    totally legimitate) edge case.

    So cletus_log's objective is to make it very easy for most apps to
    provide good logging.

    See the file "LICENSE" for the full license governing use of this file.
    Copyright 2013, 2014 Ken Farmer
"""

from __future__ import division

import os
import sys
import logging
import logging.handlers
import errno

import appdirs


class LogManager(object):

    def __init__(self,
                 app_name=None,
                 log_dir=None,
                 log_fn='main.log',
                 log_name='__main__',
                 log_file_size=100000,
                 log_count=10,
                 log_to_console=True,
                 log_to_file=True):

        self.app_name       = app_name
        self.log_name       = log_name
        self.formatter      = None
        self.log_adapter    = None
        self.logger         = logging.getLogger(log_name)
        self.log_dir        = log_dir
        self.log_fn         = log_fn
        self.log_count      = log_count
        self.log_file_size  = log_file_size

        self._create_log_formatter()

        if log_to_file:
            self._create_file_handler()

        if log_to_console:
            self._create_console_handler()

        #logging from this class isn't working
        #self.logger_sub     = logging.getLogger('%s.cletus_log' % log_name)
        #self.logger.debug('logger started, written to: %s' % self.log_dir)

        # Ensure all crashes get logged:
        sys.excepthook = self._excepthook



    def _create_log_formatter(self):
        """ Creates a formatter to ensure all records get the following format:
                <ascii time> : <name> : <level name> : <message>
            where
                <ascii time> - has a format of yyyy-mm-dd hh24:mm:ss
                <name>       - is the given module name
                <level name> - is one of DEBUG, INFO, WARNING, ERROR, CRITICAL
                <message>    - is whatever the user provided.
        """
        log_format      = '%(asctime)s : %(name)-12s : %(levelname)-8s : %(message)s'
        date_format     = '%Y-%m-%d %H.%M.%S'
        self.formatter  = logging.Formatter(log_format, date_format)



    def _create_console_handler(self):
        """ Adds a handler to send logs to the console (stdout).
        """
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(self.formatter)
        self.logger.addHandler(console_handler)



    def _create_file_handler(self):
        """ Adds a handler to send logs to a file.  This is a rotating file handler,
            so when files reach log_file_size they will get renamed and have a numeric
            suffix added.

            It will attempt to make this directory if it does not exist.
        """
        if not self.log_dir:
            if self.app_name:
                self.log_dir = appdirs.user_log_dir(self.app_name)
            else:
                print 'CRITICAL: cannot write logs to files without either log_dir or app_name'
                sys.exit(1)

        try:
            os.makedirs(self.log_dir)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                self.logger.critical('Log directory creation failed.  Log dir: %s' % self.log_dir)
                raise

        log_fqfn = os.path.join(self.log_dir, self.log_fn)

        file_handler = logging.handlers.RotatingFileHandler(log_fqfn,
                                                            maxBytes=self.log_file_size,
                                                            backupCount=self.log_count)
        file_handler.setFormatter(self.formatter)
        self.logger.addHandler(file_handler)



    def _excepthook(self, *args):
        """ Capture uncaught exceptions, write details into logger, exit.
        """
        self.logger.critical('Uncaught exception - exiting now. ', exc_info=args)
        sys.exit(1)






#!/usr/bin/env python
""" Used to ensure that only one job runs at a time.

    Cletus_job is highly opinionated about how to ensure only one job
    at a time runs, when that's important.
       - Applications should be able to wait for some amount of time for
         the older instance to complete.
       - Just checking on a pid in a file isn't enough: that pid might be
         orphaned, and you can easily get race conditions.
       - When a file system becomes full, a pidfile could be cataloged,
         but no pid can be written to it.  This needs to be handled.

    So cletus_job's objective is to make it extremely easy for most apps to
    handle this problem.

    See the file "LICENSE" for the full license governing use of this file.
    Copyright 2013, 2014 Ken Farmer
"""


import os
import errno
import time
import fcntl

import appdirs
import logging



class JobCheck(object):
    """ Ensures that only 1 job at a time runs.

    Typical Usage:
        job_check    = mod.JobCheck(pgm_name)
        if job_check.lock_pidfile():
            print 'Lock acquired, will start processing'
        else:
            print 'Try again later, already running'

        *** at end of program ***
        job_check.close()

    Inputs:
        app_name (str) - Used for two purposes: to lookup pid directory based on xdg
                         standards, and to name pid file.  Lookup only occurs if the
                         pid_dir is not provided.  Defaults to 'main'.
        log_name (str) - Should be passed from caller in order to ensure logging
                         characteristics are inherited correctly.  Defaults to __main__.
        pid_dir (str)  - Can be used instead of the automatic XDG directory
                         user_cache_dir, which is useful for testing among other things.
                         Defaults to None - which is fine as long as app_name is provided.

    Raises:
        ValueError   - Neither app_name nor pid_dir was provided, one must be
        OSError      - Could not create pid dir and it didn't already exist
    """

    def __init__(self,
                 app_name='main',
                 log_name='__main__',
                 pid_dir=None):

        self.logger   = logging.getLogger('%s.cletus_job' % log_name)
        # con't print to sys.stderr if no parent logger has been set up:
        #logging.getLogger(log_name).addHandler(logging.NullHandler())
        self.logger.debug('JobCheck starting now')

        # set up config/pidfile directory:
        self.app_name      = app_name
        self.pid_dir       = self._get_pid_dir(app_name, pid_dir)
        self.pid_fqfn      = os.path.join(self.pid_dir, '%s.pid' % self.app_name)

        self.new_pid       = os.getpid()
        self.start_time    = time.time()
        self.lock_acquired = False




    def lock_pidfile(self, wait_max=60):
        """
        Inputs:
           wait_max (int) - Maximum number of seconds for Jobcheck to keep retrying
                            lock acquisition.  If it exceeds this number it will
                            return False.
        Returns:
           True      - if lock was acquired
           False     - if lock was not acquired - pidfile is already locked
        Raises:
           IOError   - if pidfile is inaccessable
        """
        while not self._create_locked_pidfile(self.pid_fqfn, self.new_pid):
            if (time.time() - self.start_time) > wait_max:
                self.logger.error('wait_max exceeded, returning without lock')
                return False
            else:
                self.logger.warning('sleeping - waiting for lock')
                time.sleep(0.5)

        self.logger.debug('lock acquired - will return to caller')
        self.lock_acquired = True
        return True


    def _create_locked_pidfile(self, pid_fqfn, pid):
        """ Opens the pidfile, locks it, and writes the pid to it.
            Inputs:
                - pid_fqfn
                - pid
            Returns
                - True    - if locking was successful
                - False   - if locking was unsuccessful
            Raises
                - IOError - if file was inaccessible
        """
        try:
            self.pidfd = open(pid_fqfn, 'a')
        except IOError, e:
            self.logger.critical('Could not open pidfile: %s - permissions? missing dir?' % e)
            raise
        else:
            try:
                fcntl.flock(self.pidfd, fcntl.LOCK_EX|fcntl.LOCK_NB)
            except IOError:
                self.pidfd.close()
                return False
            else:
                self.pidfd.seek(0)
                self.pidfd.truncate()
                self.pidfd.write(str(pid))
                self.pidfd.flush()
                return True


    def _close_pidfile(self):
        """ Deletes pid from pidfile then closes it.
            Raises
                OSError if it cannot delete from pidfile or close it.
        """
        try:
            self.pidfd.seek(0)
            self.pidfd.truncate()
            self.pidfd.flush()
            self.pidfd.close()
        except OSError as e:
            if e.errno != errno.ENOENT:
                self.logger.critical('_delete_pidfile encountered IO error: %s' % e)
                raise


    def close(self):
        """ Final user interaction with class.
            Class can recover from prior jobs not doing this - but it's sloppy
            and could hypothetically result in errors.

        """
        if self.lock_acquired:
            self._close_pidfile()
        else:
            self.logger.warning('close() should not be called when lock was not acquired.  Will ignore.')



    def _get_pid_dir(self, app_name, arg_pid_dir):
        """Returns the pid_dir based on the arg_pid_dir if that's provided,
           otherwise by looking up the app_name in the xdg user_cache_dir.

           Inputs
                - app_name
                - arg_pid_dir 
           Returns
                - pid_dir
           Raises
                - ValueError - if neither app_name nor pid_dir is provided
                - OSError - if pid_dir doesn't exist and it can't make it
        """

        # first figure out what the directory is:
        if arg_pid_dir:
            pid_dir  = arg_pid_dir
            self.logger.debug('pid_dir will be based on arg: %s' % arg_pid_dir)
        else:
            if app_name:
                pid_dir  = os.path.join(appdirs.user_cache_dir(app_name), 'jobs')
                self.logger.debug('pid_dir will be based on user_cache_dir: %s' % pid_dir)
            else:
                err_msg = 'app_name must be provided if pid_dir is not'
                self.logger.critical(err_msg)
                raise ValueError, err_msg

        # next try to create it, just in case it isn't there
        try:
            os.makedirs(pid_dir)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                self.logger.critical('Error trying to create pid_dir: %s' % pid_dir)
                raise

        # finally, return it
        return pid_dir




#!/usr/bin/env python

import os, uuid
from pip.req import parse_requirements
from setuptools import setup, find_packages

def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

def get_version():
    v1_rec = read("cletus/_version.py")
    (v1_label, v1_version) = v1_rec.split('=')
    v2_version = v1_version.strip()[1:-1]
    assert v1_label.strip() == '__version__'
    assert v2_version.count('.') == 2
    return v2_version

VERSION          = get_version()
DESCRIPTION      = 'A library of command line utilities'
REQUIREMENTS     = [str(ir.req) for ir in parse_requirements('requirements.txt', session=uuid.uuid1())]

setup(name             = 'cletus'          ,
      version          = VERSION           ,
      description      = DESCRIPTION       ,
      long_description=(read('README.rst') + '\n\n' +
                        read('CHANGELOG.rst')),
      keywords         = "utility",
      author           = 'Ken Farmer'      ,
      author_email     = 'kenfar@gmail.com',
      url              = 'http://github.com/kenfar/cletus',
      license          = 'BSD'             ,
      classifiers=[
            'Development Status :: 4 - Beta'                         ,
            'Environment :: Console'                                 ,
            'Intended Audience :: Developers'                        ,
            'License :: OSI Approved :: BSD License'                 ,
            'Programming Language :: Python'                         ,
            'Operating System :: POSIX'                              ,
            'Topic :: Utilities'
            ],
      scripts          = ['scripts/cletus_archiver.py'],
      install_requires = REQUIREMENTS,
      packages         = find_packages(),
     )



#!/usr/bin/env  python
""" Cletus_archiver.py is a sample program to demonstrate the use of
    cletus.

    Functionally, it is extremely simple - it just compresses
    files within a directory.  Keeping it simple ensures that its
    functionality doesn't obscure the use of cletus.  And while a script
    this simple could be done with a one-line bash command in cron, this
    provides better logging, better configurability, and better
    reliability.

    Features provided by cletus include:

    - cletus_config allows the user to control the directory, the files
      chosen, and other runtime characteristics through:
        - config file
        - environmental variables
        - command line arguments

    - cletus_job allows the user to schedule the job to run
      automatically, and not worry that two may some day be running
      simultaneously with unpredictable or adverse effects.

    - cletus_log keeps the logging code simple while providing commonly
      needed features beyond the logging basic config.

    - cletus_supp allows the user to suppress the running of the program
      through a very simple mechanism, without permanently altering its
      crontab entry.

    See the file "LICENSE" for the full license governing use of this file.
    Copyright 2013, 2014 Ken Farmer
"""

#--------------------------------------------------------------


import sys
import glob
import os
import time
import argparse

import envoy

# only here to allow running out of dev
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import cletus.cletus_log    as log
import cletus.cletus_job    as job
import cletus.cletus_supp   as supp
import cletus.cletus_config as conf


logger    = None
MAX_FILES = 55
APP_NAME  = 'cletus_archiver'



def main():

    # initialization - see the functions used here for details on the
    # cletus methods.
    args       = get_args()

    setup_logging(args.log_to_console, args.log_level)
    logger.info('cletus_archiver - starting')

    config     = setup_config(args)
    logger.setLevel(config.log_level or 'DEBUG')

    jobcheck   = exit_if_already_running()

    exit_if_suppressed()

    # run the process:
    process_all_the_files()

    # housekeeping
    jobcheck.close()
    logger.info('cletus_archiver - terminating normally')
    return 0



def process_all_the_files():

    file_compressor = FileCompressor()

    for i, fn in enumerate(glob.glob('/tmp/*')):

        if i > MAX_FILES:
            logger.warning('max files exceeded - stopping now')
            break

        file_time_epoch, file_time_string = get_file_times(fn)

        if (time.time() - file_time_epoch) > (86400 * 3):

            if fn.endswith('.gz'):
                action = 'skipped'
            elif not os.path.isfile(fn):
                action = 'skipped'
            else:
                file_compressor.compress(fn)
                action = 'compressed'

        else:
            action = 'good'

        logger.debug('%-20.20s - %s - %s' % (abbreviate(fn), file_time_string, action))



class FileCompressor(object):

    def __init__(self):
        logger.debug('cletus_archiver_lib starting')

    def compress(self, fn):
        cmd   = '''gzip %s''' % fn
        r     = envoy.run(cmd)
        if r.status_code:
            logger.error('%s compression failed with status %d' % (fn, r.status_code))
        else:
            logger.debug('%s compressed' % fn)



def abbreviate(fn):
    if len(fn) > 37:
        abbrev_fn = '%s.....' % fn[:37]
        return abbrev_fn
    else:
        return fn


def get_file_times(file_name):
    time_epoch  = os.path.getatime(file_name)
    time_string = time.strftime('%Y-%m-%d %H:%M:%S %Z', time.gmtime(time_epoch))
    return time_epoch, time_string



def setup_logging(log_to_console, log_level):
    """This is a helper function to keep the bulk of the LogManager and
       these comments out of the main().

       There's nothing really special about the LogManager - it just packages
       common log configuration for you.  It includes:
          - logging to console & file by default
          - log file rotation
          - log formatting with timestamp, application name, severity level
            & message
          - an exceptionhook to log any uncaught exceptions
          - log dir creation if necessary

       LogManager accepts arguments for the log_dir, but if not provided it will
       use the app_name to look up the cache dir using the XDG standard.  On linux:
          - $HOME/.cache/<APP_NAME>/log
    """
    global logger
    cl_logger  = log.LogManager(app_name=APP_NAME,
                                log_name=__name__,
                                log_to_console=log_to_console)
    logger     = cl_logger.logger
    logger.setLevel(log_level or 'DEBUG')



def setup_config(args):
    """This is a helper function to keep the bulk of the ConfigManager and
       these comments out of the main().

       ConfigManager is being used to first add config items from a file,
       then from the argparse namespace.   Alternatively, we could add:
          - multiple config files (eg, server config plus this app config)
          - environmental variables
          - arguments & options from optparse, etc that uses dictionaries
            instead of namespaces
          - config items from a resource like zookeeper that are loaded
            via the add_iterable ConfigManger method.

       The order determines the precidence.  Since args are added after the
       config file any matching items will result in args overriding the
       config file.

       Validation is performed using validictory.  This is optional, but
       recommended.  All config items are stored in two places:  one
       dedicated to that method, and one consolidated dictionary:
          - cm_config_iterable
          - cm_config_file
          - cm_config_namespace
          - cm_config_env
          - cm_config           - the consolidated config
       Any of these dictionaries can be used for validation, just pass in
       its name (just drop the cm_ prefix) for the config_type argument.
       And you might want to provide a different schema for different configs,
       just pass that it using the schema argument.

       More info on Validictory is here:
          - https://readthedocs.org/projects/validictory/
       Note that any validation tool can be easily used with the dictionaries.
    """
    config_schema = {'type': 'object',
                     'properties': {
                       'dir':       {'required': True,
                                     'type':     'string'},
                       'log_level': {'enum': ['DEBUG','INFO','WARNING','ERROR','CRITICAL']} ,
                       'config_fqfn': {'required': False,
                                       'type':     'string'},
                       'log_to_console': {'required': False,
                                          'type':     'boolean'} },
                     'additionalProperties': False
                    }
    config = conf.ConfigManager(config_schema)
    config.add_file(app_name=APP_NAME,
                    config_fqfn=args.config_fqfn,
                    config_fn='main.yml')
    config.add_namespace(args)
    config.validate()
    return config




def exit_if_already_running():
    """This is a helper function to keep the bulk of the cletus_job check, and
       these comments out of main().

       Because it's providing the APP_NAME, JobCheck will use the XDG standard
       for the location of the cache file.  On linux this would be:
            $HOME/.cache/<APP_NAME>/jobs
       Alternatively, we could have passed in a specific pid_dir, which would
       get used instead of the XDG dir.

       Since we didn't pass in a log_name, it will default to __main__.

       lock_pidfile defaults to 60 seconds for wait time, checking every 0.5
       seconds.  We could pass in the number of seconds to this function if we
       want to override that.
    """
    jobcheck  = job.JobCheck(app_name=APP_NAME)
    if jobcheck.lock_pidfile():
        logger.info('JobCheck has passed')
        return jobcheck
    else:
        logger.warning('Pgm is already running - this instance will be terminated')
        sys.exit(0)



def exit_if_suppressed():
    """This is a helper function to keep the bulk of this check out of main().
       It runs the cletus_supp SuppressCheck class and provides it with the application
       name defined above.   Since no config_dir was provided, SuppressCheck will look
       for a config file in the xdg standard location.  On linux this will be:
            $HOME/.config/<APP_NAME>/suppress

       Every time suppcheck.suppressed is called it will look for suppression files
       within the suppression directory.  Since no specific suppression_name was
       provided as an argument, it will default to using the app_name with this
       convention:
            $HOME/.config/<APP_NAME>/suppress/name-<APP_NAME>.suppress

       Note that this function could be called at checkpoints by this program, in
       order to identify that it should shut-down gracefully in the middle of long-
       running processes.
    """

    suppcheck  = supp.SuppressCheck(app_name=APP_NAME)
    if suppcheck.suppressed():
        logger.warning('Pgm has been suppressed - this instance will be terminated.')
        sys.exit(0)
    else:
        logger.info('SuppCheck has passed')




def get_args():
    """" Returns all options & arguments for program.

         Uses argparse, but could alternatively use optparse, getopt, docopt, etc.

         Notice that validation is occuring within this process - and this validation
         will be done again using the ConfigManager and Validictory.  Validation in
         either location is optional, and this is redundant.  However, there are some
         benefits to doing it here as well - mostly in the form of better usability.
    """

    usage  = """Cletus sample command-line application that demonstrates
                how to use the cletus libraries.

                This application will compress old files.
             """
    parser = argparse.ArgumentParser(description=usage)

    parser.add_argument('--log-level',
                        choices=['debug','info','warning','error', 'critical'])
    parser.add_argument('--config-fqfn')
    parser.add_argument('--console-log',
                        action='store_true',
                        default=True,
                        dest='log_to_console')
    parser.add_argument('--no-console-log',
                        action='store_false',
                        dest='log_to_console')
    args = parser.parse_args()

    return args



if __name__ == '__main__':
    sys.exit(main())


